/* $Id: nalloc.c 3178 2005-02-27 22:42:07Z bremner $
 * 
 * Node allocation library
 * David Bremner, 1999/04/01
 *
 * a "node" is an area of memory, within which we will do memory allocation
 * The idea is to allow a set of several (variable sized) structures to be
 * stored in a contiguous piece of memory.
 * 
 * To recover a given struct, one needs to know which "slot" it is in.
 * If p is a point to an  na_node_t, then p->slot[k] contains an offset into
 * p->mem where struct k is stored.
 * 
 * the macro na (in nalloc.h) can be used to access a variable.
 * see znode.h for example declarations.
 *
 * The allocation model is rather primitive; the assumption is 
 * that all memory in a given node is allocated once, and never deallocated.
 * The node will grow to fit.
 * 
 * There are some hairy casting things going on here; the hope is that 
 * by aligning things on sizeof(long) boundaries, we avoid alignment 
 * problems.
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stddef.h>
#ifdef HAVE_CONFIG_H
#include "ssds_config.h"
#endif
#include "ssds_debug.h"
#include "nalloc.h"
#include "error_message.h"
#ifdef USE_PROFILE
#include "profile.h"
#endif
#include "memalloc.h"
#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

#define CHECKANDRETURN(p) { NODECHECK(p) ; return(p); }

static void na_desc_init(na_desc_p dp){
  dp->in_use=FALSE;
  dp->dangles=FALSE;
  dp->offset=-1;
  dp->ptr=NULL;
  dp->fixup_tag=-1;
  dp->destroy_tag=-1;
  dp->wipe_aux_tag=-1;
  dp->free_aux_tag=-1;
  dp->report_tag=-1;
}


/*
 * create a new node
 * This must be called first.
 */

na_node_p na_new(MINT tag){
  return na_new1(tag,NA_INITIAL_SIZE);
}

na_node_p na_new1(MINT tag, size_t size){
  na_node_p p;
  int i;
#ifdef USE_PROFILE
  prof_event(PROF_CREATE_NODE);
#endif

  err_debug(SD_MEMTRACE,  (HERE,"allocating %ld",size));
  p=ZALLOC(sizeof(na_node_t)+size,tag);

  err_assert(offsetof(na_node_t,mem)%sizeof(long)==0);

  p->frozen=1;
  p->destroyed=0;
  for (i=0; i< NA_SLOTS; i++){
    na_desc_init(p->chunk+i);
  }

  p->size=0;
  p->limit=size;

  CHECKANDRETURN(p);

}

/** 
 * This routine allocates the space within a node.
 * The pointer returned is _NOT_ to the memory allocated; it is 
 * to the (possibly memreallocated) node structure.
 */   
na_node_p na_alloc(na_node_p p, na_slot_t slot,size_t size){
  size_t needed;


  needed=ALIGN(size);

  err_assert(p->chunk[slot].in_use==0);
  err_assert(size>=0);

  p->chunk[slot].in_use=1;

  if (slot >= NA_SLOTS) 
    err_fatal(HERE,"slot %d out of range", slot);
      

  if (p->size+needed > p->limit)
    { 
      size_t new_limit=p->limit*NA_ALLOC_MUL+needed+NA_ALLOC_INCR;
      err_debug(SD_MEMTRACE,  (HERE,"memreallocating %ld",size));
      p=ZREALLOC(p,sizeof(na_node_t)+new_limit);
      if (!p){
	err_fatal("na_alloc","memrealloc to %ld bytes failed", 
		  sizeof(na_node_t)+new_limit);
      }
      p->limit=new_limit;
    } 

  p->chunk[slot].offset=p->size;
  p->chunk[slot].size=needed;
  p->chunk[slot].fixup_tag=-1;
  p->chunk[slot].wipe_aux_tag=-1;
  p->chunk[slot].destroy_tag=-1;
  p->chunk[slot].report_tag=-1;
  p->size+=needed;
  CHECKANDRETURN(p);
}




/* print out statistics for a given node */
void na_report(na_node_p p){
  int i;
  NODECHECK(p);

  fprintf(stderr,"Total node allocation %dk\n",p->limit/1024);
  for (i=0; i< NA_SLOTS; i++){
      if (p->chunk[i].in_use){
	fprintf(stderr,"chunk %d size %d\n",
		i,p->chunk[i].size);
	if (p->chunk[i].report_tag>=0){
	  na_jump_table[p->chunk[i].report_tag](p,i);
	}

      }
  }
}


/**
 * Destroy a node, invoking any destroy hooks on data structures 
 * in that node

 * @param p pointer to node to be destroyed.
 */

void na_destroy(na_node_p p){
  int i;
  NODECHECK(p);

#ifdef USE_PROFILE
  prof_event(PROF_DESTROY_NODE);
#endif

  for (i=0; i< NA_SLOTS; i++){
      if (p->chunk[i].in_use && p->chunk[i].destroy_tag>=0){
	na_jump_table[p->chunk[i].destroy_tag](p,i);
	p->chunk[i].in_use=0;
      }
  }

  p->destroyed=1;
#ifdef USE_ZRAM
  z_FreeNode((z_NodeP)p); 
#else
  free(p);
#endif
}

/*
 * Wipe (set to null) any auxilary pointers.
 * This is so shallow copies do not share auxilary structures/

*/
void na_wipe_aux(na_node_p p){
  int i;
  NODECHECK(p);

  for (i=0; i< NA_SLOTS; i++){
      if (p->chunk[i].in_use && p->chunk[i].wipe_aux_tag>=0){
	na_jump_table[p->chunk[i].wipe_aux_tag](p,i);
      }
  }
}

/**
   free any auxilary data structures
*/

void na_free_aux(na_node_p p){
  int i;
  NODECHECK(p);

  for (i=0; i< NA_SLOTS; i++){
      if (p->chunk[i].in_use && p->chunk[i].free_aux_tag>=0){
	na_jump_table[p->chunk[i].free_aux_tag](p,i);
      }
  }
}


/* This gives how many bytes to e.g. ship across a network */
size_t na_size(na_node_p p){
  return (p->size+sizeof(na_node_t));
}

void *na_access(na_node_p p, na_slot_t i){

  err_assert(p!=NULL);
  if (  !p->chunk[i].in_use){
    err_fatal(HERE,"chunk %d is not in use",i);
  }

  if (p->chunk[i].dangles){ 
    return (p->chunk[i].ptr);
  } else {
    return (void *)(p->mem+p->chunk[i].offset);
  }
}

static void na_copy_tags(na_node_p dest, na_node_p src, slot_t i){
    
  dest->chunk[i].fixup_tag=  src->chunk[i].fixup_tag;
  dest->chunk[i].destroy_tag=  src->chunk[i].destroy_tag;
  dest->chunk[i].wipe_aux_tag=  src->chunk[i].wipe_aux_tag;
  dest->chunk[i].free_aux_tag=  src->chunk[i].free_aux_tag;
  dest->chunk[i].report_tag=src->chunk[i].report_tag;
}
/** na_freeze will pack any dangling chunks back into the node.
   Currently there is no attempt at making this fast.

   @todo this should make its mind up about whether or not the 
   parameter is preserved. At the moment, auxilary pointers (
   e.g. sarray stacks) are shallow copied.  

   @param p na_node. 
 */

na_node_p na_freeze(na_node_p p){
  
  
  int i;
  na_node_p new_node;

  NODECHECK(p);

  err_fatal(HERE,"don't call me!");

  new_node=na_new1(p->tag,p->size);


  
  for(i=0; i< NA_SLOTS; i++){
    if (p->chunk[i].in_use){
      size_t bytes=p->chunk[i].size;
      new_node=na_alloc(new_node,i,bytes);
      
      na_copy_tags(new_node,p,i);
      if (p->chunk[i].dangles){
	memcpy(na_access(new_node,i),p->chunk[i].ptr, bytes);
      } else {
	memcpy(na_access(new_node,i),na_access(p,i),bytes);
      }
    }
  }
  
  na_fixup(new_node);
  new_node->frozen=1;

  CHECKANDRETURN(new_node);

}

/** 
 * Make a copy of a node suitable for external transport.
 * 
 * Any auxilary data is discarded.
 * @param p node
 * 
 * @return new node
 */

na_node_p na_compact(na_node_p p){
  
  
  int i;
  na_node_p new_node;

  NODECHECK(p);
  
  new_node=na_new1(p->tag,p->size);
  
  for(i=0; i< NA_SLOTS; i++){
    if (p->chunk[i].in_use){
      size_t bytes=p->chunk[i].size;
      new_node=na_alloc(new_node,i,bytes);
      
      na_copy_tags(new_node,p,i);
      if (p->chunk[i].dangles){
	memcpy(na_access(new_node,i),p->chunk[i].ptr, bytes);
      } else {
	memcpy(na_access(new_node,i),na_access(p,i),bytes);
      }
      if (new_node->chunk[i].wipe_aux_tag >= 0)
	na_jump_table[new_node->chunk[i].wipe_aux_tag](new_node,i);
    }
  }
  
  na_fixup(new_node);
  new_node->frozen=1;

  CHECKANDRETURN(new_node);

}



void na_dump_header(na_node_p p){
  int i;

  err_assert(p!=NULL);

  for(i=0;i<NA_SLOTS;i++){

    na_desc_p dp=p->chunk+i;
    
    fprintf(stderr,"chunk[%d]\toff %d size %d in_use %d fixup tag %d\n",
	   i,dp->offset,dp->size,dp->in_use,dp->fixup_tag);


  }
}

void na_sanity_check(char* loc, na_node_p p){
  int i;

  err_assert1(loc,p!=NULL);
  err_assert1(loc,!p->destroyed);
#ifdef USE_ZRAM
  err_assert1(loc,p->child != (z_NodeP)p);
#endif
  for(i=0;i<NA_SLOTS;i++){

    na_desc_p dp=p->chunk+i;
    
    err_assert1(loc,dp->in_use==0 || dp->size>0);
   
    if (i>0){
      na_desc_p dp1=p->chunk+i-1;

      err_assert1(loc,(dp->in_use==0 ||  (dp->offset !=dp1->offset)));
    }
  }
}

/* this does a simple memcpy, except that it preserves 
   the ZRAM_HEADER portion of dest
*/ 
void na_copy(na_node_p dest, na_node_p src){
  int i;
  dest->size=src->size;
  dest->limit=src->limit;

  for (i=0; i< NA_SLOTS; i++){
    dest->chunk[i]=src->chunk[i];
  }

  (void)memcpy(dest->mem,src->mem,src->size);
  NODECHECK(dest);
}


void *na_realloc(na_node_p p, na_slot_t slot, size_t new_size){

  na_desc_p dp;

  NODECHECK(p);
  err_debug(SD_MEMTRACE,  (HERE,"reallocating %ld",new_size));
  dp=p->chunk+slot;

  err_assert(dp->in_use); /* there should be something there first */

  if (dp->dangles){
    /* already dangling */
    dp->ptr=memrealloc(dp->ptr,new_size);
    err_assert(dp->ptr);
    dp->size=new_size;
  } else {
    
    p->frozen=0;
    
    dp->dangles=1;
    dp->ptr=memalloc(new_size);
    memcpy(dp->ptr,p->mem+dp->offset,dp->size);
    dp->size=new_size;
  }
  NODECHECK(p);
  return (dp->ptr);

}

/* Fix up any pointers trashed by copying */

void na_fixup(na_node_p p){

  int i;

  NODECHECK(p);

  for (i=0; i<NA_SLOTS; i++){
    if (p->chunk[i].in_use){
      if (p->chunk[i].fixup_tag >=0){
	na_jump_table[p->chunk[i].fixup_tag](p,i);
      }
    }
  }
}

void na_set_fixup(na_node_p p,na_slot_t i,tag_t tag){
  p->chunk[i].fixup_tag=tag;
}

void na_set_report(na_node_p p,na_slot_t i,tag_t tag){
  p->chunk[i].report_tag=tag;
}

void na_set_wipe_aux(na_node_p p,na_slot_t i,tag_t tag){
  p->chunk[i].wipe_aux_tag=tag;
}

void na_set_free_aux(na_node_p p,na_slot_t i,tag_t tag){
  p->chunk[i].free_aux_tag=tag;
}

void na_set_destroy(na_node_p p,na_slot_t i,tag_t tag){
  p->chunk[i].destroy_tag=tag;
}
