/* $Id: basis_equation.c 3182 2005-03-01 17:15:38Z bremner $ */

/** 
 *  This file provides functions for creating and manipulating
 * <em>basis equation blocks</em>.

 *  Conceptually, a <em>basis equation</em>
 * is a list of signed basis indices.
 * NOTE: this file now contains only the generic data structure functions.
 * specific kinds of constraints are now treated in pluecker.c and boundary.c
 */

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_CONFIG_H
#include "config.h"
#endif
#include "oms_opt.h"
#include "oms_hook.h"
#include "combinat.h"
#include "be_internal.h"
#include "int_queue.h"
#include "statearray.h"
#include "nalloc.h"
#include "globaldef.h"
#include "nodedef.h"
#include "int_queue.h"
#include "sa16_array.h"
#include "profile.h"
#include "bitmask.h"
#include "bindex_table.h"
#include "nu_sign_set.h"
#include "debug.h"

#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

TAG_DECLARE(be_fixup);

static void be_preinit(be_block_p new);


/**
 * Allocate a basis equation block.
 */

extern be_block_p be_alloc(size_t rows, 
			   size_t cols, 
			   size_t bases){
  be_block_p new;

  new=malloc(sizeof(be_block_t));
  ERR_TRAP_NULL(new);

  new->rows=rows;
  new->cols=cols;
  new->bases=bases;

  new->last=malloc(bases*sizeof(array_index_t));
  ERR_TRAP_NULL(new->last);

  new->row_mask=malloc(rows*sizeof(bitmask_t));
  ERR_TRAP_NULL(new->row_mask);

  new->eqn=malloc(rows*cols*sizeof(be_coeff_t));
  ERR_TRAP_NULL(new->eqn);

  be_preinit(new);

  return (new);
}

static void be_preinit(be_block_p new){
  int i;
  for (i=0; i< new->rows; i++){
    int j;
    for (j=0; j<new->cols; j++){
      be_mark_unused(new,i,j);
      new->eqn[i*(new->cols)+j].prev=-1;
    }
  }

}
/**
   Fix any pointers munged in transit.
*/
extern void be_fixup(na_node_p node,array_index_t slot){
  be_block_p be;
  
  be=na_access(node,slot);
  be->last=na_access(node,RSLOT(slot,be_block,last));
  be->row_mask=na_access(node,RSLOT(slot,be_block,row_mask));
  be->eqn=na_access(node,RSLOT(slot,be_block,eqn));
}
/**
 Allocate a basis equation block within a node.
*/
extern na_node_p be_nalloc(na_node_p node,
			   array_index_t slot,
			   size_t rows, 
			   size_t cols, 
			   size_t bases){
  be_block_p new;

  node=na_alloc(node,slot,sizeof(be_block_t));
  node=na_alloc(node,RSLOT(slot,be_block,last),bases*sizeof(array_index_t));
  node=na_alloc(node,RSLOT(slot,be_block,row_mask),rows*sizeof(bitmask_t));
  node=na_alloc(node,RSLOT(slot,be_block,eqn),rows*cols*sizeof(be_coeff_t));

  na_set_fixup(node,slot,TAG(be_fixup));
  new=na_access(node,slot);
  ERR_TRAP_NULL(new);

  new->rows=rows;
  new->cols=cols;
  new->bases=bases;

  be_fixup(node,slot);
  new->initialized=0;

  be_preinit(new);
  return (node);
}


extern array_index_t be_bases(be_block_p BE){
  return BE->bases;
}

extern array_index_t be_cols(be_block_p BE){
  return BE->cols;
}

extern array_index_t be_rows(be_block_p BE){
  return BE->rows;
}
extern array_index_t be_bindex(be_block_p BE, array_index_t row, array_index_t col){
  err_assert(row < BE->rows);
  err_assert(col < BE-> cols);
  
  return be_bindex_raw(BE,row*(BE->cols)+col);
}

extern array_index_t be_sign(be_block_p BE, array_index_t row, array_index_t col){
  err_assert(row < BE->rows);
  err_assert(col < BE-> cols);
  
  return be_sign_raw(BE,row*(BE->cols)+col);
}

extern void be_set_bindex(be_block_p BE, array_index_t row, array_index_t col,array_index_t val){

  int array_index_t=row*(BE->cols)+col;

  err_assert(row < BE->rows);
  err_assert(col < BE-> cols);

#ifdef BE_STORE_SIGN  
  BE->eqn[array_index_t].bindex=val;
#else
  {
    
    int cur_sign;
    if (be_in_use(BE,array_index_t)){
      cur_sign=be_sign_raw(BE,array_index_t);
    } else {
      cur_sign=1;
    }
    BE->eqn[array_index_t].sindex=si_encode(val,cur_sign);
  }
#endif 
}

extern void be_set_sign(be_block_p BE, array_index_t row, array_index_t col,array_index_t val){
  int array_index_t=row*(BE->cols)+col;

  err_assert(row < BE->rows);
  err_assert(col < BE-> cols);


#ifdef BE_STORE_SIGN  
  BE->eqn[array_index_t].sign=val;
#else
  {
    int cur_bindex;
    if (be_in_use(BE,array_index_t)){
      cur_bindex=be_bindex_raw(BE,array_index_t);
    } else {
      cur_bindex=0;
    }


    BE->eqn[array_index_t].sindex=si_encode(cur_bindex,val);
  }
#endif 
}


/** Compute the linked lists that allow us to find all of the
 * equations referencing a given basis.
 */

extern void be_makelinks(be_block_p be){
  int i,j;

  be->ones_mask= (1 << be->cols)-1;
  for (i=0; i< be->bases; i++){
    be->last[i]=-1;
  }
  

  for (i=0; i< be->rows; i++){
    for (j=0; j< be->cols; j++){
      array_index_t array_index_t,b;

      array_index_t=i*be->cols+j;

      b=be_bindex_raw(be,array_index_t);

      err_assert((b >=-1) && (b<be->bases));
      if (b>=0){
	be->eqn[array_index_t].prev=be->last[b];

#ifdef BE_STORE_ROW
	be->eqn[array_index_t].row=i;
#endif
#ifdef BE_STORE_ROW_START
	be->eqn[array_index_t].row_start=i*be->cols;
#endif
#ifdef BE_STORE_COL
	be->eqn[array_index_t].col=j;
#endif
	be->last[b]=array_index_t;
      }
    } 
  }
  be->initialized=1;
}


/**
   Debugging routine
*/
extern void be_dump(be_block_p BE){
  array_index_t i,j;


  for (i=0 ; i< BE->rows; i++){
    printf("[%d:%x]\t", i,BE->row_mask[i]);
    for (j=0; j< BE->cols; j++){
      printf("%c%d\t", be_sign(BE,i,j) > 0 ? ' ' : '-' ,be_bindex(BE,i,j));
    }
    printf("\n");
  }   
}

/** 
 * 
 * 
 * @param mask_array array of masks, one per row
 * @param which sign to test for presence of
 * @param row mask row to possibly update
 * @param col column
 * @param sigma set of signs to check
 * 
 * @return updated  mask
 */

nu_sign_t be_update_mask(sarray_p mask_array, nu_sign_t which, 
		      array_index_t row, array_index_t col,
		      nu_sign_t sigma){
  bitmask_t mask;

  if (sign_subset(which,sigma)){
    mask=sa_set_bits(mask_array,row,1,col,1);
  } else { 
    mask=sa_set_bits(mask_array,row,0,col,1);
  }

  return mask;
}

			  
void be_check_mask(na_node_p global,
		   be_block_p BE, array_index_t row,
		   bitmask_t mask, nu_sign_t sigma,
		   sarray_p chi){
  int j;

  bitmask_t cur_mask=mask;
  params_p params=na_access(global,SLOT_params);
  nu_sign_t initial;

  if (params->uniform){
    initial=SIGN_NONZERO;
  } else {
    initial=SIGN_ANY;
  }

  for (j=0; j<be_cols(BE) && (be_bindex(BE,row,j)>=0); j++){
    array_index_t basis=be_bindex(BE,row,j);
    int eq_sign=be_sign(BE,row,j);
    nu_sign_t chi_basis=sa_read(chi,basis);
    nu_sign_t current=sign_mulint(chi_basis,eq_sign);


    
    bitmask_t bit=cur_mask&BIT0_MASK;

    if ((bit && !sign_subset(sigma,current) ) || 
	(!bit && (current != initial) && sign_subset(sigma,current)))
      err_fatal("be_check_mask","mask %x, array_index_t %d does not match %d*chi[%d]=%s",
		mask,j,eq_sign,basis,sign_name[current]);
    cur_mask >>= 1;

  }
}

/** 
 * 
 * 
 * @param BE basis equations
 * @param elem pseudopointer to element
 * @param sigma 
 * @param pmask mask of positive positions
 * @param nmask mask of negative positions
 * @param zmask mask of zero positions
 * @param forced set of signs to force
 * 
 * @return true if row contains at least one positive and one negative sign
 */
bool be_nonmonotone_row(be_block_p BE,
			array_index_t elem,
			nu_sign_t sigma,
			bitmask_t pmask, 
			bitmask_t nmask, 
			bitmask_t zmask,
			nu_sign_set_p forced){

  bitmask_t missing_bit,row_mask;
  array_index_t row;

  row=be_row(BE,elem);
  row_mask=BE->row_mask[row];

  err_debug(DB_BOUNDARY,("be_nonmonotone_row", "pmask %x nmask %x zmask %x row_mask %x ",
		       pmask,nmask,zmask,row_mask));
  if ( ((pmask|nmask|zmask)&row_mask)==row_mask){
    /* all signs determined */
    if ((pmask==0) || (nmask==0)){
      /* oops, looks like a facet */
      return(FALSE);
    }
  } else {   
    missing_bit= (~(pmask|nmask|zmask))&row_mask;
    if ( ZERO_OR_POWER_OF_TWO(missing_bit)  && ((pmask==0) || (nmask==0))){
      nu_sign_t new_sign;
      array_index_t basis,array_index_t;
	    
      err_debug(DB_BOUNDARY,("be_nonmonotone_row", "pm %x nm %x zm %x mb %x ",
		       pmask,nmask,zmask,missing_bit));
	    
      err_assert(missing_bit && ZERO_OR_POWER_OF_TWO(missing_bit));

	    
      array_index_t=be_row_start(BE,elem)+LOG2(missing_bit);
      basis=be_bindex_raw(BE,array_index_t);
      
      if (pmask==0)
	new_sign=SIGN_POS;
      else
	new_sign=SIGN_NEG;

      new_sign=sign_mulint(new_sign,be_sign_raw(BE,array_index_t));
	    

      err_debug(DB_BOUNDARY|DB_FORCE,
		(HERE, "nonmonotone row %d forcing chi[%d] to %s",
		 be_row(BE,elem), basis,sign_name[new_sign]));

      if (!nss_add(forced,basis,new_sign)) return FALSE;
    }
  }
  return TRUE;
}


