/* pmrs_main.c
   Enumeration of polyominoes by reverse search
   using connected induced subgraphs.

   The root of the tree is a one-point graph. The forward search function
   removes one point of the graph such that it stays connected.

   A backtrack solution is more efficient.
   Ambros Marzetta

   size	solutions	time [s] (DEC 3000-M700 Alpha)
   2	1
   3	2
   4	5
   5	12
   6	35
   7	108
   8	369
   9	1285		2
   10	4655		8
   11	17073		43
   12	63600		173
   13			907
   
 * $Source$
 * $Revision: 2732 $
 * $Date: 2000-11-15 12:14:52 -0400 (Wed, 15 Nov 2000) $
 * $Log$
 * Revision 1.2  2000/11/15 16:14:52  bremner
 * committing patched version of zram source
 *
 * Revision 1.1  1998/06/17  18:17:42  marzetta
 * Initial revision
 *
 */

#include <stdlib.h>		/* exit() */
#include <stdio.h>
#include "z_io.h"		/* getopt() */
#include "z_time.h"
#include "z_enumps.h"
#include "cis_base.h"
#include "z_signal.h"
#include "pm_both.h"

static z_rs_ProcSet pm_procset1, pm_procset2, pm_procset3;
static int pm_procsetnum1, pm_procsetnum2, pm_procsetnum3;

typedef struct pm_Global {	/* subclass of cis_Global */
  struct cis_Global pm_cis;
  int pm_n, pm_h, pm_w;
} pm_Global, *pm_GlobalP;

#define pm_coord_x(gp, i)	ps2d_coord_x(gp->pm_cis.cis_coord, i)
#define pm_coord_y(gp, i)	ps2d_coord_y(gp->pm_cis.cis_coord, i)
#define pm_adj(global)		adjlist_adj(global->pm_cis.cis_adjlist)


/* Functions needed for reverse search ----------------------------------- */

boolean pm_Full(pm_GlobalP global, cis_NodeP n)
{
  /* Return true if the polyomino has the requested number of tiles and
     is lexicographically greatest of symmetric polyominoes. */
  int x[MAXN], y[MAXN], a[MAXN];
  int i, count, xmin, xmax, ymin, ymax, xx, yy;

  if (n->numAct != global->pm_n) return 0;
  
  count = 0;
  xmin = ymin = MAXN * 3;
  xmax = ymax = -1;
  for (i=global->pm_cis.numPt-1; i>=0; --i) /* find min/max coordinates */
    if (n->pt[i]) {
      xx = i % global->pm_w;
      yy = i / global->pm_w;
      if (xx < xmin) xmin = xx;
      if (xx > xmax) xmax = xx;
      if (yy < ymin) ymin = yy;
      if (yy > ymax) ymax = yy;
      x[count] = xx;
      y[count] = yy;
      count ++;
    }
  assert(ymax == global->pm_h-1);
  assert(count == global->pm_n);
  for (i=count-1; i>=0; --i) {
    x[i] -= xmin;		/* translate */
    y[i] -= ymin;
    a[i] = x[i] * MAXN + y[i];	/* store points in array a */
  }
  xmax -= xmin;
  ymax -= ymin;
  return lexmax(a, count, xmax, ymax);
} /* pm_Full */

boolean pm_is_root_node(pm_GlobalP global, cis_NodeP this)
{
  return this->numAct == 1;
} /* pm_is_root_node */

cis_NodeP pm_adj_oracle(pm_GlobalP global, cis_NodeP node, uint32 k)
{
  /* Limit the search tree: Use connected induced subgraph oracle if
     the subgraph has not enough tiles. */
  if (node->numAct < global->pm_n)
    return cis_adj_oracle(&global->pm_cis, node, k);
  else
    return NULL;
} /* pm_adj_oracle */


/* Postscript output ------------------------------------------------------ */

void pm_PSHeader(pm_GlobalP global, FILE *psfile, float xsize, float ysize)
{
  float side;

  side = ysize / global->pm_h;
  fprintf(psfile, "/side %7.2f def\n", side);
  fprintf(psfile, "/square {\n");
  fprintf(psfile, "  newpath side mul exch side mul exch moveto\n");
  fprintf(psfile, "  side 0 rlineto 0 side rlineto\n");
  fprintf(psfile, "  0 side sub 0 rlineto closepath\n");
  fprintf(psfile, "  gsave 0.5 setgray fill grestore stroke} def\n");
  fprintf(psfile, "\n");
} /* pm_PSHeader */

void FindMin(pm_GlobalP global, cis_NodeP node, int *xmin, int *ymin)
     /* called by pm_PSPicture */
{
  int pt, x, y;
  
  *xmin = *ymin = global->pm_w;
  for (pt=0; pt<global->pm_cis.numPt; pt++) {
    if (node->pt[pt]) {
      x = pt % global->pm_w;
      y = pt / global->pm_w;
      if (x < *xmin) *xmin = x;
      if (y < *ymin) *ymin = y;
    } /* if pt */
  } /* for pt */
} /* FindMin */

void pm_PSPicture(pm_GlobalP global, cis_NodeP node, FILE *psfile,
		  float xsize, float ysize)
{
  int pt, xmin, ymin, x, y;
  
  assert(global && global->pm_cis.tag == cis_GlobalTag);
  fprintf(psfile, "\n");
  FindMin(global, node, &xmin, &ymin);
  for (pt=0; pt<global->pm_cis.numPt; pt++) {
    if (node->pt[pt]) {
      x = pt % global->pm_w - xmin;
      y = pt / global->pm_w - ymin;
      fprintf(psfile, "%d %d square\n", x, y);
    } /* if pt */
  } /* for pt */
} /* pm_PSPicture */


/* Character graphic output ----------------------------------------------- */

#define CHARPERLINE 80

void MakeLine(pm_GlobalP global, cis_NodeP *n, int k)
{
  int i, j, line, pt;
  int xmin[CHARPERLINE / 3], ymin[CHARPERLINE / 3];
  boolean fill;
  
  for (i=0; i<k; i++) FindMin(global, n[i], &xmin[i], &ymin[i]);
  
  for (line=0; line<global->pm_h; line++) {
    for (i=0; i<k; i++) {	/* element i */
      for (j=0; j<global->pm_n; j++) { /* character j */
	pt = line * global->pm_w + xmin[i] + j;
	fill = (pt < global->pm_cis.numPt && n[i]->pt[pt]);
	printf("%c", fill ? '#' : ' ');
      }
      if (i < k-1) printf("  "); /* space between elements */
    }
    printf("\n");
  }
  printf("\n");			/* empty line between element rows */
  
  for (i=0; i<k; i++) z_FreeNode((z_NodeP)n[i]);
} /* MakeLine */

void MakeCharacterGraphic(pm_GlobalP global)
{
  int32 elem_count;
  int elemPerLine, i, k;
  cis_NodeP n[CHARPERLINE / 3];
  
  z_enumps_Rewind(&elem_count);
  elemPerLine = CHARPERLINE / (global->pm_n+2);
  k = 0;
  for (i=0; i<elem_count; i++) {
    n[k++] = (cis_NodeP) z_Load();
    if (k == elemPerLine) {
      MakeLine(global, n, k);
      k = 0;
    }
  }
  if (k) MakeLine(global, n, k);
} /* MakeCharacterGraphic */


/* Initialization -------------------------------------------------------- */

void initproc(void)
{
  cis_base_Init();
  z_rs_Init();
  z_enumps_init();
  z_signal_Init();
  
  pm_procset1.rs_Equal = (z_EqualProc)cis_Equal;
  pm_procset1.rs_Condition = (z_ConditionProc)pm_Full;
  pm_procset1.rs_Result = (z_NodeProc)cis_Print;
  pm_procset1.rs_F = (z_rs_FProc)cis_f;
  pm_procset1.rs_AdjOracle = (z_rs_AdjOracleProc)pm_adj_oracle;
  pm_procset1.rs_IsRoot = (z_ConditionProc)pm_is_root_node;
  pm_procset1.rs_Delta = (z_rs_DeltaProc)cis_Delta;
  pm_procset1.rs_FindChildNumber = NULL;
  pm_procset1.rs_Forward = NULL;
  pm_procset1.rs_NextChild = NULL;
  pm_procsetnum1 = z_InstallProcSet((z_ProcSetP) &pm_procset1);
  
  pm_procset2 = pm_procset1;	/* character graphic / Postscript */
  pm_procset2.rs_Result = z_Save;
  pm_procsetnum2 = z_InstallProcSet((z_ProcSetP) &pm_procset2);
  
  pm_procset3 = pm_procset1;	/* no output */
  pm_procset3.rs_Result = NULL;
  pm_procsetnum3 = z_InstallProcSet((z_ProcSetP) &pm_procset3);
} /* initproc */


/* Main program ----------------------------------------------------------- */

static int Usage(const char *name)
{
  fprintf(stderr, "Usage: %s [-1chnprt] n\n\n", name);
  fprintf(stderr, "Options: -1  sequential\n");
  fprintf(stderr, "         -c  character graphic output\n");
  fprintf(stderr, "         -h  help (produce this message)\n");
  fprintf(stderr, "         -n  no output (implies time measurement)\n");
  fprintf(stderr, "         -p  PostScript output\n");
  fprintf(stderr, "         -r  restart\n");
  fprintf(stderr, "         -t  time measurement\n");
  return 4;
} /* Usage */

int mmain(int argc, char **argv, char **envp)
{
  pm_GlobalP global;
  cis_NodeP root;
  int i, j, n, w, h, numpt;
  double time;
  unsigned int index;
  enum {ot_normal, characterGraphic, postScript, noOutput}
       outputType = ot_normal;
  boolean showTime=0;
  boolean sequential=0;
  boolean restart=0;
  int procsetnum;
  char c;

  while ((c=getopt(argc, argv, "1chnprt")) != -1)
    switch(c) {
    case 'c':
      outputType = characterGraphic;
      break;
    case 'n':
      outputType = noOutput;
      showTime = 1;
      break;
    case 'r':
      restart = 1;
      break;
    case 'p':
      outputType = postScript;
      sequential = 1;
      break;
    case 't':
      showTime = 1;
      break;
    case '1':
      sequential = 1;
      break;
    default:
      return Usage(argv[0]);
    }
  
  if (argc != optind+1) return Usage(argv[0]);
  
  n = atoi(argv[optind]);
  if (n < 2) {
    fprintf(stderr, "n must be >= 2\n");
    return 3;
  }
  
  h = (n+1) / 2;
  w = 2 * n - 2;
  numpt = h*w-n+2;
  if (numpt > MAXDEGREE) {
    fprintf(stderr, "%d is too large for n\n", n);
    exit(5);
  }
  global = (pm_GlobalP) z_NewNode(cis_GlobalTag, sizeof(pm_Global));
  if (!global) exit(1);
  global->pm_n = n;
  global->pm_h = h;
  global->pm_w = w;
  global->pm_cis.numPt = h*w-n+2;
  global->pm_cis.cis_adjlist = z_adjlist_New(numpt, w*h*4);
  global->pm_cis.cis_coord = z_ps2d_New(numpt);
  index = 0;
  for (i=0; i<h; i++)
    for (j=0; j<w; j++)
      if (i*w+j < numpt) {
	adjlist_first(global->pm_cis.cis_adjlist)[i*w+j] = index;
	if (i != 0)   pm_adj(global)[index++] = i*w+j-w;
	if (i != h-1 && i*w+j+w < numpt)
	  pm_adj(global)[index++] = i*w+j+w;
	if (j != 0)   pm_adj(global)[index++] = i*w+j-1;
	if (j != w-1 && i*w+j+1 < numpt)
	  pm_adj(global)[index++] = i*w+j+1;
	pm_adj(global)[index++] = ADJ_LAST;
	pm_coord_x(global, i*w+j) = j;
	pm_coord_y(global, i*w+j) = i;
      }
  if (! restart) {
    root = cis_EmptySubgraph(&global->pm_cis);
    root->numAct = 1;
    root->pt[(h-1)*w+n-1] = 1;
  }
  
  switch (outputType) {
  case noOutput:
    procsetnum = pm_procsetnum3; /* no output */
    break;
  case characterGraphic:
  case postScript:
    procsetnum = pm_procsetnum2; /* save */
    z_enumps_Open("/tmp/pm");
    break;
  default:
    procsetnum = pm_procsetnum1; /* print */
    break;
  }
  
  time = z_Gtime();
  if (restart)
    z_rs_Restart(procsetnum, (z_lb_GlobalP) global, sequential);
  else
    z_rs_Enumerate(procsetnum, (z_lb_GlobalP) global,
		   (z_NodeP)root, sequential);
  time = z_Gtime() - time;

  if (outputType == characterGraphic)
    MakeCharacterGraphic(global);
  else if (outputType == postScript) {
    z_enumps_ratio = 0.5;
    z_enumps_frame = 0;
    z_enumps_borderFraction = 1/18.0;
    z_PrintPS((z_NodeP)global, stdout,
	      (void*)pm_PSHeader, (void*)pm_PSPicture);
  }
  if (showTime)
    if (sequential)
      printf("Time:%7.2f (sequential) n = %d\n",
	     time, n);
    else
      printf("Time:%7.2f (%2d processor%s n = %d\n",
	     time, z_numprocs, z_numprocs>1?"s)":") ", n);
  return 0;
} /* mmain */
