/* Reverse search interface example for ZRAM
   Ambros Marzetta

   Enumerate all elements of a rectangular grid by reverse search. The
   origin (0,0) of the grid is in the south-west corner. A grid element
   has at most four neighbours. The local search function goes west if
   possible, otherwise south.
   
 * $Source$
 * $Revision: 2732 $
 * $Date: 2000-11-15 12:14:52 -0400 (Wed, 15 Nov 2000) $
 * $Log$
 * Revision 1.2  2000/11/15 16:14:52  bremner
 * committing patched version of zram source
 *
 * Revision 1.1  1996/06/03  19:00:19  marzetta
 * Initial revision
 *
 *
 */

#include <stdio.h>
#include "z_base.h"
#include "z_rs.h"

typedef struct grid_Node {
  Z_NODE_HEADER;		/* necessary for all objects */
  int x,y;
} grid_Node, *grid_NodeP;

typedef struct grid_Global {
  Z_NODE_HEADER;		/* necessary for all objects */
  int w,h;			/* width and height of grid */
} grid_Global, *grid_GlobalP;

/* Static variables (initialized on every processor by initproc() -- */

static int grid_NodeTag, grid_GlobalTag; /* for z_InstallClass */
static int grid_procsetnum; /* for z_InstallProcSet */
static z_rs_ProcSet grid_procset; /* pointers to application-specific
				     procedures */

/* Mandatory application-specific procedures ----------------------- */

boolean grid_Equal(grid_GlobalP global, grid_NodeP a, grid_NodeP b)
{
  assert(a);
  assert(b);
  return (a->x == b->x) && (a->y == b->y);
} /* grid_Equal */

uint32 grid_Delta(grid_GlobalP global, grid_NodeP node)
{
  return 4;	       /* a grid element has at most four neighbours */
} /* grid_Delta */

grid_NodeP grid_AdjOracle(grid_GlobalP global, grid_NodeP node, uint32 k)
{
  grid_NodeP result;

  result = NULL;
  switch (k) {
  case 0:			/* north */
    if (node->y < global->h-1) {
      result = (grid_NodeP) z_NewCopyNode((z_NodeP) node);
      result->y++;
    }
    break;
  case 1:			/* east */
    if (node->x < global->w-1) {
      result = (grid_NodeP) z_NewCopyNode((z_NodeP) node);
      result->x++;
    }
    break;
  case 2:			/* south */
    if (node->y > 0) {
      result = (grid_NodeP) z_NewCopyNode((z_NodeP) node);
      result->y--;
    }
    break;
  case 3:			/* west */
    if (node->x > 0) {
      result = (grid_NodeP) z_NewCopyNode((z_NodeP) node);
      result->x--;
    }
    break;
  }
  return result;
} /* grid_AdjOracle */

boolean grid_IsRoot(grid_GlobalP global, grid_NodeP node)
{
  /* return true on the origin where grid_F is not defined. */
  
  return (node->x == 0) && (node->y == 0);
} /* grid_IsRoot */

grid_NodeP grid_F(grid_GlobalP global, grid_NodeP node)
{
  grid_NodeP result;

  assert(node);
  assert((node->x > 0) || (node->y > 0));
  if (node->x > 0) {		/* go west */
    result = (grid_NodeP) z_NewCopyNode((z_NodeP) node);
    result->x--;
    return result;
  }
  result = (grid_NodeP) z_NewCopyNode((z_NodeP) node);
  result->y--;			/* go south */
  return result;
} /* grid_F */

void grid_Print(grid_GlobalP global, grid_NodeP node)
{
  printf("[%2d](%d,%d)\n", z_myproc, node->x, node->y);
} /* grid_Print */


/* Optional application-specific procedures ------------------------ */

#ifdef MORE_EFFICIENT
uint32 grid_FindChildNumber(grid_GlobalP global, grid_NodeP child,
			    grid_NodeP parent)
/* Inverse function to the adjacency oracle: Return the first (and unique)
   integer k such that Adj(parent, k) == child. */
{
  if (child->y == parent->y+1) return 0;
  if (child->x == parent->x+1) return 1;
  if (child->y == parent->y-1) return 2;
  assert(child->x == parent->x-1);
  return 3;
} /* grid_FindChildNumber */

void grid_Forward(grid_GlobalP global, grid_NodeP *nodeH, uint32 *k)
/* This corresponds to a sequence of F, FindChildNumber, FreeNode */
{
  assert(*nodeH);
  assert(((*nodeH)->x > 0) || ((*nodeH)->y > 0));
  if ((*nodeH)->x > 0) {	/* go west */
    (*nodeH)->x--;
    *k = 1;
  } else {
    (*nodeH)->y--;		/* go south */
    *k = 0;
  }
} /* grid_Forward */
#endif /* MORE_EFFICIENT */

/* Processor initialization and main program ----------------------- */

void initproc(void)
{
  /* This function is called once on every processing element. */

  z_rs_Init();

  grid_NodeTag = z_InstallClass(grid_Node, 0, 0, 0);
  grid_GlobalTag = z_InstallClass(grid_Global, 0, 0, 0);

  grid_procset.rs_Equal = (z_EqualProc) grid_Equal;
  grid_procset.rs_F = (z_rs_FProc) grid_F;
  grid_procset.rs_AdjOracle = (z_rs_AdjOracleProc) grid_AdjOracle;
  grid_procset.rs_IsRoot = (z_ConditionProc) grid_IsRoot;
  grid_procset.rs_Delta = (z_rs_DeltaProc) grid_Delta;
  grid_procset.rs_Result = (z_NodeProc) grid_Print;
#ifdef MORE_EFFICIENT
  grid_procset.rs_FindChildNumber = (z_rs_FindChildNumberProc)
                                    grid_FindChildNumber;
  grid_procset.rs_Forward = (z_rs_ForwardProc) grid_Forward;
#endif /* MORE_EFFICIENT */
  
  grid_procsetnum = z_InstallProcSet((z_ProcSetP) &grid_procset);
} /* initproc */

int mmain(int argc, char **argv, char **envp)
{
  /* main program */
  grid_GlobalP global;
  grid_NodeP root;

  global = (grid_GlobalP) z_NewNode(grid_GlobalTag, sizeof(grid_Global));
  global->w = 4;
  global->h = 5;

  root = (grid_NodeP) z_NewNode(grid_NodeTag, sizeof(grid_Node));
  root->x = 0;
  root->y = 0;

  z_rs_Enumerate(grid_procsetnum, (z_lb_GlobalP) global, (z_NodeP) root, 0);
  return 0;
} /* mmain */
