/* z_bb2.c (ZRAM: Tree size estimator)
   Ambros Marzetta

 * $Source$
 * $Revision: 2732 $
 * $Date: 2000-11-15 12:14:52 -0400 (Wed, 15 Nov 2000) $
 * $Log$
 * Revision 1.2  2000/11/15 16:14:52  bremner
 * committing patched version of zram source
 *
 * Revision 1.6  1997/04/25  08:41:11  marzetta
 * Added some type casts.
 *
 * Revision 1.5  1996/07/22  11:41:50  marzetta
 * Fixed estimator bug by using double.
 *
 * Revision 1.4  1996/05/29  17:30:10  marzetta
 * Sequential version implemented.
 *
 * Revision 1.3  1996/04/22  13:52:56  marzetta
 * Fixed memory leak.
 *
 * Revision 1.2  1996/04/19  14:15:46  marzetta
 * Fixed bug in error and time estimates.
 *
 * Revision 1.1  1996/04/09  13:15:08  marzetta
 * Initial revision
 *
 */

#include <math.h>		/* sqrt */
#include "z_bb.h"
#include "z_par.h"
#include "s_seq.h"
#include "z_time.h"
#include <stdio.h>
#include <stdlib.h>


/* Static variables ------------------------------------------------------ */

static boolean bb2_inited;
static double sum, sum2;
static long maxBranch;

#ifndef SEQUENTIAL
static z_JobNum z_bb_estimateJob;
#endif /* SEQUENTIAL */


/* Estimator ------------------------------------------------------------- */

struct glob {
  s_seqP seq;
  int numChildren, numSolutions;
  z_NodeP upperBound;
  z_bb_ProcSetP psp;
};
  
void DoEstimateChild(z_NodeP child, struct glob *ref)
{
  /* This procedure is given to psp->bb_Branch as whatToDo */
  assert(child);
  /* Ignore children above the upper bound */
  if (ref->psp->bb_CompareProc(child, ref->upperBound) <= 0) {
    if (ref->psp->bb_IsSolution(child)) {
      ref->numSolutions ++;
      z_FreeNode(child);
    } else {
      ref->numChildren ++;
      s_seq_InsertLast(ref->seq, child);
    }
  } else
    z_FreeNode(child);
} /* DoEstimateChild */

void DoEstimateChildInPlace(z_NodeP child, struct glob *ref)
{
  /* This procedure is given to psp->bb_BranchInPlace as whatToDo */
  assert(child);
  /* Ignore children above the upper bound */
  if (ref->psp->bb_CompareProc(child, ref->upperBound) <= 0) {
    if (ref->psp->bb_IsSolution(child))
      ref->numSolutions ++;
    else {
      ref->numChildren ++;
      s_seq_InsertLast(ref->seq, z_NewCopyNode(child));
    }
  }
} /* DoEstimateChildInPlace */

static void Estimate(z_bb_ProcSetP psp, z_bb_GlobalDataP global,
	      z_NodeP root, z_NodeP ub,
	      long numEstimates, double *sum, double *sum2, long *branch)
{
  /* Find numEstimates paths through the search tree. */
  int i, j, select;
  z_NodeP node;
  double factor, count;
  s_seq_refP sref;
  struct glob glob;

  glob.psp = psp;
  glob.upperBound = ub;
  glob.seq = s_seq_New(s_seq_default);
  *sum = *sum2 = 0;
  *branch = 0;			/* number of branches for time measurement */
  for (i=numEstimates; i>0; --i) {
    factor = 1;
    count = 0;
    node = z_NewCopyNode(root);	/* start a new path */
    for (;;) {
      count += factor;		/* count current node */
      glob.numSolutions = glob.numChildren = 0;
      if (psp->bb_Branch)
	psp->bb_Branch(global, node, ub,
		       (z_bb_DoProcP) DoEstimateChild, &glob);
      else
	psp->bb_BranchInPlace(global, node, ub,
			      (z_bb_DoProcP) DoEstimateChildInPlace, &glob);
      (*branch)++;
#ifdef ESTIMATE_NOISY
      printf("numChildren: %d, factor: %5.1f %s\n",
	     glob.numChildren, factor, psp->bb_ShortNodeString(node));
#endif /* ESTIMATE_NOISY */
      z_FreeNode(node);
      count += factor * glob.numSolutions;
      if (! glob.numChildren) break;	/* end of path */
      factor *= glob.numChildren;
      select = rand() % glob.numChildren; /* select a random child */
      for (j = select; j > 0; --j) {
	sref = s_seq_First(glob.seq);
	z_FreeNode((z_NodeP) s_seq_GetObject(glob.seq, sref));
	s_seq_Delete(glob.seq, sref);
      }
      sref = s_seq_First(glob.seq);
      node = (z_NodeP) s_seq_GetObject(glob.seq, sref);
      s_seq_Delete(glob.seq, sref);
      while (s_seq_numElements(glob.seq)) {
	sref = s_seq_First(glob.seq);
	z_FreeNode((z_NodeP) s_seq_GetObject(glob.seq, sref));
	s_seq_Delete(glob.seq, sref);
      }
    } /* for (;;) */
    *sum += count;
    *sum2 += count * count;
#ifdef ESTIMATE_NOISY
    printf("count = %7.1f\n", count);
#endif /* ESTIMATE_NOISY */
  } /* numEstimates */
  s_seq_Free(glob.seq);
} /* Estimate */


/* Parallelism related procedures ---------------------------------------- */

#ifndef SEQUENTIAL

static void z_bb_estimate(void)
{
  z_ProcSetNum psn;
  z_bb_ProcSetP psp;
  z_NodeP root, upperBound;
  z_bb_GlobalDataP global;
  long numEstimates, d;
  double sum, sum2;
  long branch;
  
  {
    /* Read the message broadcast by the front end in z_BranchBoundEstimate.
       It contains global data */
    psn = z_GetMINTBuffer();
    global = (z_bb_GlobalDataP) z_GetNodeBuffer();
    root = z_GetNodeBuffer();
    upperBound = z_GetNodeBuffer();
    numEstimates = z_GetMINTBuffer();
    psp = (z_bb_ProcSetP) z_GetProcSet(psn);
  }
  
  d = numEstimates % z_numprocs;
  numEstimates /= z_numprocs;
  if (z_myproc < d) numEstimates ++;
  /*printf("[%2d] %d estimates\n", z_myproc, numEstimates);*/

  Estimate(psp, global, root, upperBound, numEstimates, &sum, &sum2, &branch);
  
  z_FreeNode(root);
  z_FreeNode(global);
  z_FreeNode(upperBound);
  z_AddDataBuffer(&sum, sizeof(sum));
  z_AddDataBuffer(&sum2, sizeof(sum2));
  z_AddDataBuffer(&branch, sizeof(branch));

} /* z_bb_estimate */

void z_EstimateResult(int pe)
{
  /* This procedure is called on the front end and read the results of the
     parallel estimates from the input buffer */
  double f;
  long branch;

  z_GetDataBuffer(&f, sizeof(f));	sum += f;
  z_GetDataBuffer(&f, sizeof(f));	sum2 += f;
  z_GetDataBuffer(&branch, sizeof(branch));
  if (branch > maxBranch) maxBranch = branch;
} /* z_EstimateResult */

#endif /* ! SEQUENTIAL */


/* Main procedure ------------------------------------------------------- */

void z_BranchBoundEstimate(z_ProcSetNum procsetnum,
			   z_bb_GlobalDataP global,
			   z_NodeP root, z_NodeP UpperBound,
			   int numEstimates,
			   float *nNodes, float *sNodes,
			   float *nTime, float *sTime)
/* Estimate the number of nodes with a value less than UpperBound in the
   search tree. Evaluate numEstimates random paths through the tree.
   Return an estimate for the number of nodes and the (sequential) time
   together with the standard deviation of the estimate. */
{
  z_bb_ProcSetP psp;
  double time;
  
  assert(z_frontend);		/* can be called on the front end only */
  assert(bb2_inited);
  assert(numEstimates > 0);
  
  psp = (z_bb_ProcSetP) z_GetProcSet(procsetnum);
  assert(psp->bb_Branch || psp->bb_BranchInPlace);
  assert(psp->bb_CompareProc);
  assert(psp->bb_IsSolution);
  if (psp->bb_IsSolution(root)) {
    *nNodes = 1;
    *sNodes = 0;
    *nTime = 0;
    *sTime = 0;
    return;
  }
  time = z_Gtime();
  
#ifndef SEQUENTIAL
  /* Broadcast: procsetnum, *global, *root */
  {
    z_AddMINTBuffer(procsetnum);
    z_AddNodeBuffer(global);
    z_AddNodeBuffer(root);
    z_AddNodeBuffer(UpperBound);
    z_AddMINTBuffer(numEstimates);
    z_StartParallelJob(z_bb_estimateJob, z_EstimateResult);
  }

  /* Wait / work */
  z_WaitWork();
#else /* SEQUENTIAL */
  Estimate(psp, global, root, UpperBound, numEstimates,
	   &sum, &sum2, &maxBranch);
#endif /* SEQUENTIAL */
  
  time = z_Gtime() - time;
  /*printf("Result: %5.1f %5.1f %5.1f %d\n", sum, sum2, 0.0, numEstimates);*/
  *nNodes = sum / numEstimates;
  if (numEstimates > 1)
    *sNodes = sqrt((sum2 - sum*sum/numEstimates) / (numEstimates - 1)
		   / numEstimates);
  else *sNodes = 0;
  /*printf("time = %5.1f\n", time);*/
  *nTime = time / maxBranch * *nNodes;
  *sTime = time / maxBranch * *sNodes;
} /* z_BranchBoundEstimate */


/* Initialization -------------------------------------------------------- */

void z_bb2_Init(void)
{
  /* called on every processor by initproc() */

  if (! bb2_inited) {
    bb2_inited = 1;
    srand(z_myproc * z_numprocs);
#ifndef SEQUENTIAL
    z_bb_estimateJob = z_InstallJob(z_bb_estimate);
#endif /* SEQUENTIAL */
  }
} /* z_bb2_Init */
