/* z_bb4.c (redistribute checkpoint files for z_BranchBoundDFExp)
   Ambros Marzetta

   This program strongly depends on the checkpoint file format defined in
   z_lb.c and z_bb3.c.

   Proposistion for command line interface:
   Either  app -k nin nout  or  app -r nin.

 * $Source$
 * $Revision: 2732 $
 * $Date: 2000-11-15 12:14:52 -0400 (Wed, 15 Nov 2000) $
 * $Log$
 * Revision 1.2  2000/11/15 16:14:52  bremner
 * committing patched version of zram source
 *
 * Revision 1.1  1996/05/15  16:55:39  marzetta
 * Initial revision
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include "z_bb.h"
#include "z_lb.h"		/* z_lb_CheckpointPath */
#include "z_io.h"
#include "z_ps.h"
#include "s_seq.h"


/* Constants, Types from z_lb.c ----------------------------------------- */

#define CHECKPOINT_VERSION 4
#define CHECKPOINT_END 0xA4A4A4A4

typedef enum StatEnumT {stat_w, stat_s, stat_ws, stat_or, stat_rf,
			stat_last} StatEnumT;
typedef long StatArrayT[stat_last];


/* Static variables ----------------------------------------------------- */

static StatArrayT stat;
static s_seqP workList;
static int cp_sequence;
static z_NodeP globalUpperBound;
static boolean bb4_inited;


/* Procedures ----------------------------------------------------------- */

static void PrintStatistics(void)
{
  printf("Total WorkAndSplit %ld\n", stat[stat_ws]);
  printf("Total Requests generated %ld\n", stat[stat_or]);
  printf("Total Requests forwarded %ld\n", stat[stat_rf]);
  fflush(stdout);
} /* PrintStatistics */

int RenameAll(int nin)
     /* Rename nin files check.* to check.*.bak */
     /* return 1 if one of them doesn't exist */
{
  char from[PATH_MAX], to[PATH_MAX];
  int result;
  int i;

  for (i=0; i<nin; i++) {
    z_io_AppendProc(from, z_lb_CheckpointPath, nin, i);
    strcpy(to, from);
    strcat(to, ".bak");
    result = rename(from, to);
    if (result) {
      perror(from);
      return 1;
    }
  }
  return 0;
} /* RenameAll */

int ReadAll(z_bb_ProcSetP psp, int nin)
{
  char from[PATH_MAX];
  int result;
  int i, j, s, numprocs;
  FILE *file;
  int version, num, checkpoint_end, cp_seq;
  z_NodeP thisUpperBound;
  StatArrayT lstat;
  
  for (i=0; i<nin; i++) {
    z_io_AppendProc(from, z_lb_CheckpointPath, nin, i);
    strcat(from, ".bak");
    file = fopen(from, "r");
    if (!file) {
      perror(from);
      return 2;
    }
    z_fread(&version, sizeof(version), 1, file);
    if (version < 2 || version > 4) {
      fprintf(stderr, "*%2d*bad checkpoint file version\n", i);
      return 8;
    }
    if (version >= 4) {
      z_fread(&numprocs, sizeof(numprocs), 1, file);
      if (numprocs != nin) {
	fprintf(stderr, "*%2d*checkpoint file is for %d processors\n",
		i, numprocs);
	return 7;
      }
      z_fread(&cp_seq, sizeof(cp_seq), 1, file);
      if (!i)
	cp_sequence = cp_seq;
      else
	if (cp_sequence != cp_seq) {
	  fprintf(stderr,
		  "*%2d*checkpoint file has sequence number "
		  "%d instead of %d\n",
		  i, cp_seq, cp_sequence);
	  return 7;
	}
    } /* version >= 4 */
    z_fread(&lstat, sizeof(lstat), 1, file);
    for (s=0; s<stat_last; s++) stat[s] += lstat[s];
    z_fread(&num, sizeof(num), 1, file);
    fprintf(stderr, "[%2d]in: Nodes %d\n", i, num);
    for (j=0; j<num; j++)
      s_seq_InsertLast(workList, z_ReadNode(file));
    thisUpperBound = z_ReadNode(file);
    if (thisUpperBound) {
      if (! globalUpperBound) globalUpperBound = thisUpperBound;
      else {
	if (psp->bb_CompareProc(thisUpperBound, globalUpperBound) < 0) {
	  z_FreeNode(globalUpperBound);
	  globalUpperBound = thisUpperBound;
	} else
	    z_FreeNode(thisUpperBound);
      }
    }
    if (version >= 3) {
      z_fread(&checkpoint_end, sizeof(checkpoint_end), 1, file);
      if (checkpoint_end != CHECKPOINT_END) {
	fprintf(stderr, "*%2d*bad checkpoint file end\n", z_myproc);
	return 9;
      }
    }
    fclose(file);
  } /* for */
  if (psp->bb_ShortNodeString)
    printf("Current upper bound: (%s)\n",
	   psp->bb_ShortNodeString(globalUpperBound));
  return 0;
} /* ReadAll */

static void WriteCheckpointHeader(FILE *cp_file,
				  int version, int cp_sequence, int numprocs,
				  StatArrayT *stat, int num)
{
  fwrite(&version, sizeof(version), 1, cp_file);
  fwrite(&numprocs, sizeof(numprocs), 1, cp_file);
  fwrite(&cp_sequence, sizeof(cp_sequence), 1, cp_file);
  fwrite(stat, sizeof(StatArrayT), 1, cp_file);
  fwrite(&num, sizeof(num), 1, cp_file);
} /* WriteCheckpointHeader */

void WriteAll(int nout, int numElements)
{
  char name[PATH_MAX];
  int i, j, numPerFile, numThisFile, numprocs = nout,
    version = CHECKPOINT_VERSION, checkpoint_end = CHECKPOINT_END;
  FILE *cp_file;
  static StatArrayT lstat;
  z_NodeP workUnit;
  s_seq_refP ref;

  cp_sequence ++;
  numPerFile = numElements / nout;
  for (i=0; i<nout; i++) {
    numThisFile = numPerFile;
    if (i < numElements % nout) numThisFile ++;
    z_io_AppendProc(name, z_lb_CheckpointPath, nout, i);
    cp_file = fopen(name, "w");
    WriteCheckpointHeader(cp_file, version, cp_sequence, numprocs,
			  &lstat, numThisFile);
    fprintf(stderr, "[%2d]out: Nodes %d\n", i, numThisFile);
    for (j=0; j<numThisFile; j++) {
      ref = s_seq_First(workList);
      workUnit = s_seq_GetObject(workList, ref);
      z_WriteNode(cp_file, workUnit);
      s_seq_Delete(workList, ref);
      z_FreeNode(workUnit);
    }
    z_WriteNode(cp_file, globalUpperBound);
    fwrite(&checkpoint_end, sizeof(checkpoint_end), 1, cp_file);
    fclose(cp_file);
  } /* for i */
} /* WriteAll */

int z_bb_RedistributeDFExp(z_ProcSetNum procsetnum, int n_in, int n_out)
/* Read checkpoint files created by n_in processors and write them as
   if they were created by n_out processors. Print the statistical data
   and reinitialize it to zero. */
{
  int res;
  z_bb_ProcSetP psp;

  assert(bb4_inited);
  psp = (z_bb_ProcSetP) z_GetProcSet(procsetnum);
  assert(psp->bb_CompareProc);
  if (n_in < 1) {
    fprintf(stderr, "z_bb_RedistributeDFExp: n_in must be > 0\n");
    return 6;
  }
  if (n_out < 1) {
    fprintf(stderr, "z_bb_RedistributeDFExp: n_out must be > 0\n");
    return 6;
  }
  res = RenameAll(n_in);
  if (res) return res;
  res = ReadAll(psp, n_in);
  if (res) return res;
  PrintStatistics();
  WriteAll(n_out, s_seq_numElements(workList));
  return 0;
} /* z_bb_RedistributeDFExp */

void z_bb4_Init(void)
{
  /* called on every processor by initproc() */
  if (!bb4_inited) {
    bb4_inited = 1;
    z_io_Init();

    workList = s_seq_New(s_seq_default);
  }
} /* z_bb4_Init */

/* Issues:

   Dependency on checkpoint file format in z_lb.c (z_bb3.c is inevitable).

   */
